#!/usr/bin/perl

# This script maintains the popauth.db file for dovecot.

# It looks into its spool directory for <ip>-<pid> files, which the
# mail_executable (/usr/lib/dovecot/dovecot-log.sh) is supposed to create here
# If one is found, existence of pid is checked. If it still exists, pidfile
# is touched to signify process is still around. If no such pid is there, file
# is removed if older than grace period

# TODO: daemonize
#

use strict;
use Getopt::Long;
use English;
use Fcntl qw(:DEFAULT :flock :seek);
use DB_File;
use Net::Netmask;
use POSIX 'setsid';
use Linux::Inotify2;
use POSIX;
use Errno;
use Time::HiRes qw(usleep);

# Default grace period is 30*60 seconds
my $grace=30*60;
my $spoolDir="/var/spool/popauth";
my $dbfile="/etc/mail/popauth";
my $pid_file = '/var/run/popauth.pid';
my $mta_pid_file = '/var/run/mta/sendmail.pid';
my $ignoreNets = '127.0.0.0/8';
my $doKill;

use Sys::Syslog qw(:DEFAULT setlogsock);
# Try uncommenting this if you get no syslog output:
#setlogsock('unix');  # Valid arg values: inet, unix, stream, console.
openlog('popauth', 'pid', 'mail');


sub log_msg {
    my ($level, $msg) = @_;
    syslog($level,$msg);
#    print STDERR "$msg\n";
}

sub usage {
    die "Usage: $0 [-g grace] [-d spooldir] [-f dbFile] [-p sendmailPidFile] [-i ignNet1/ignMask1,ignNet2/ignMask2]\n";
}


# ---------------------------------------------
# Following code snarfed from PopBeforeSmtp conf
my %db;
my $signal_sendmail = 0;
my $sendmail_pid;
my $flock=1;

sub get_sendmail_pid
{
    if(!open(PID, $mta_pid_file)) { 
	log_msg('error', "Unable to open $mta_pid_file: $!");
	return 0;
    }			
    $_ = <PID>;
    my($pid) = /(\d+)/;
    close PID;
    return $pid;
}

my $dbh;

# We set the global %db to the opened database hash.  We also set $dbh for
# our sync function, and DB_FH for our flock_DB function.
sub tie_sendmail
{
    $dbh = tie %db, 'DB_File', "$dbfile.db", O_CREAT|O_RDWR, 0666, $DB_HASH
	or die "$0: cannot dbopen $dbfile: $!\n";
    if ($flock) {
	my $fd = $dbh->fd;
	open(DB_FH,"+<&=$fd") or die "$0: cannot open $dbfile filehandle: $!\n";
    }
}

sub sync_sendmail
{
    $dbh->sync and die "$0: sync $dbfile: $!\n";

    while ($signal_sendmail && 
	   ($sendmail_pid <= 0 || !kill(1, $sendmail_pid))) {
	my $new_pid = get_sendmail_pid();
	if ($new_pid == $sendmail_pid) {
	     log_msg('error', 
		     "Unable to signal sendmail to reread the database.\n");
	     return;
	}
	$sendmail_pid = $new_pid;
    }
}

# Only 1 of these 2 locking functions is used (look for $flock_func above).
sub flock_sendmail
{
    my($locking) = @_;
    flock(DB_FH, $locking ? LOCK_EX : LOCK_UN)
	or die "$0: flock_sendmail($locking) failed: $!\n";
}

sub fcntl_sendmail
{
    my($locking) = @_;
    my $lock = pack('s s l l i', $locking ? F_WRLCK : F_UNLCK, SEEK_SET, 0, 0, 0);
    fcntl(DB_FH, F_SETLKW, $lock)
	or die "$0: fcntl_sendmail($locking) failed: $!\n";
}

my $tie_func = \&tie_sendmail;
my $sync_func = \&sync_sendmail;
#my $flock_func = \&flock_sendmail;
my $flock_func = \&fcntl_sendmail;

if ($signal_sendmail) {
    $sendmail_pid = get_sendmail_pid();
}

# End code copied from popBeforeSmtp
# ---------------------------------------------

my $nextWake; # When the process needs to wake up next

sub schedWake {
    my ($wake) = @_;
    if(!defined $nextWake || $wake < $nextWake) {
	$nextWake = $wake;
    }
}

sub parse {
    my ($old) = @_;
    my %ip;
    my $now=time();
    my $haveAlive=0; # Do we have processes that are alive?
    my $changed=0;
    opendir(DIR, $spoolDir) || die "Could not open $spoolDir ($ERRNO)\n";
    while(my $file = readdir(DIR)) {
	my $fullName="$spoolDir/$file";
	if($file eq "stop") {
	    log_msg("info", "exiting");
	    unlink $fullName;
	    exit(0);
	}
	    
	next if($file !~ /^(\d+\.\d+\.\d+\.\d+)-(\d+)/); # Bad pattern => ignore
	my ($ip, $pid) = ($1, $2);

	if(defined findOuterNetblock($ip)) {
	    log_msg("info", "$ip in ignored list, removing");
	    unlink($fullName);
	    next;
	}
	
	# Check whether that pid still exists...
	my $r = kill 0, $pid;
	if($r == 0) {
	    # TODO: Distinguish other error causes than ESRSCH...
	    
	    # Process does not exist
	    #log_msg("debug", "Process $pid gone");
	    my (@stat) = lstat($fullName);
	    my $wake=$stat[9] + $grace;
	    if($wake > $now) {
		# File still recent, keep it
		$ip{$ip}=1;
		schedWake($wake);
	    } else {
		unlink($fullName);
	    }
	} else {
	    #log_msg("debug", "Process $pid still present, touching file");
	    utime(undef, undef, $fullName) ||
		log_msg("error", "Could not touch $fullName ($ERRNO)\n");
	    $ip{$ip}=1;
	    $haveAlive=1;
	}
    }
    close(DIR);
    if($haveAlive) {
	# We have alive processes. Schedule wakeup after $grace / 2
	schedWake($now + $grace/2);
    }

    if(defined $old) {
	foreach my $i (keys %ip) {
	    if(!exists $old->{$i}) {
		# A new ip
		$changed=1;
	    }
	}
	foreach my $i (keys %$old) {
	    if(!exists $ip{$i}) {
		# Removed ip
		$changed=1;
	    }
	}
    }
    return ($changed,\%ip);
}

sub sync {
    my ($new) = @_;
    my $changed=0;
    $flock_func->(1);
    # Add missing
    foreach my $ip (keys %$new) {
	if(!exists $db{$ip}) {
	    $db{$ip}="ok";
	    log_msg("info", "Adding $ip");
	    $changed=1;
	}
    }
    foreach my $ip (keys %db) {
	if(!exists $new->{$ip}) {
	    delete $db{$ip};
	    log_msg("info", "Deleting $ip");
	    $changed=1;
	}
    }
    if($changed) {
	$sync_func->();
    }
    $flock_func->(0);
}


my $ret = GetOptions("grace|g=i" => \$grace,
		     "spooldir|d=s" => \$spoolDir,
		     "db|f=s" => \$dbfile,
		     "mtaPidFile|m=s" => \$mta_pid_file,
		     "pidFile|p=s" => \$pid_file,
		     "ignoreNets|i=s" => \$ignoreNets,
		     "kill|k" => \$doKill);
if(!$ret) {
    usage();
}

if($ignoreNets) {
    for (split(",",$ignoreNets)) {
	my $blk = Net::Netmask->new2($_);
	if(!$blk) {
	    die "Bad netblock $_\n";
	}
	$blk->storeNetblock();
    }
}

sub check_process {
    my ($pid) = @_;
    my $r=kill(0, $pid);
    if($r) {
	return 1; # Process alive
    }
    return 0 if $! == Errno::ESRCH;
    die "Can't probe process $pid\n";
}

# Pid file
if(-f $pid_file) {
    open(PID, "<", $pid_file) || die "Could not open pid file ($ERRNO)\n";
    my $prev_pid=<PID>;
    chomp($prev_pid);
    close(PID);
    if($prev_pid < 2) {
	die "Bad prev pid $prev_pid\n";
    }
    if($doKill) {
	kill(SIGTERM, $prev_pid) || die "Could not kill $prev_pid ($ERRNO)\n";
	for(my $i=0; $i<10; $i++) {
	    if(!check_process($prev_pid)) {
		exit(0);
	    }
	    usleep(60000);
	}
	kill(SIGKILL, $prev_pid);
	for(my $i=0; $i<10; $i++) {
	    if(!check_process($prev_pid)) {
		exit(0);
	    }
	    usleep(60000);
	}
	exit(1);
    }
    if(check_process($prev_pid)) {
	die "Process already running\n";
    }
}

if($doKill) {
    die "Process not running\n";
}

unlink("$spoolDir/stop");
$tie_func->();

# First shot
my ($changed,$ip)=parse();
sync($ip);


sub daemonize {
#    chdir '/'               or die "Can't chdir to /: $!";
    open STDIN, '/dev/null' or die "Can't read /dev/null: $!";
    open STDOUT, '>/dev/null'
	or die "Can't write to /dev/null: $!";
    defined(my $pid = fork) or die "Can't fork: $!";
    exit if $pid;
    setsid                  or die "Can't start a new session: $!";
    open STDERR, '>&STDOUT' or die "Can't dup stdout: $!";
}

daemonize();

open(PID, ">", $pid_file);
print PID "$$\n";
close(PID);

sub term {
    unlink($pid_file);
    exit(0);
}

sub alrm {
}

$SIG{'ALRM'}=\&alrm;
$SIG{'TERM'}=\&term;

my $inotify = new Linux::Inotify2();
$inotify->watch($spoolDir, IN_CREATE);
while(1) {
    if(defined $nextWake) {
	my $now=time();
	if($nextWake > $now+1) {	    
	    alarm($nextWake-$now);
	}
    }
    my @events = $inotify->read();
    alarm(0);
#    foreach (@events)
#    {
#	log_msg("debug", sprintf "File: %s; Mask: %d\n", $_->fullname, $_->mask);
#    }
    undef $nextWake;
    ($changed,$ip)=parse($ip);
#    log_msg("Changed=$changed");
    if($changed) {
	sync($ip);
    }
}
